import unittest

from yatl.helpers import DIV, SPAN, TAG, XML


class TestHelpers(unittest.TestCase):
    def test_tags(self):
        DIV = TAG.div
        IMG = TAG["img/"]
        self.assertEqual(DIV().xml(), "<div></div>")
        self.assertEqual(IMG().xml(), "<img/>")
        self.assertEqual(DIV(_id="my_id").xml(), '<div id="my_id"></div>')
        self.assertEqual(IMG(_src="crazy").xml(), '<img src="crazy"/>')
        self.assertEqual(
            DIV(_class="my_class", _mytrueattr=True).xml(),
            '<div class="my_class" mytrueattr="mytrueattr"></div>',
        )
        self.assertEqual(
            DIV(
                _id="my_id", _none=None, _false=False, without_underline="serius?"
            ).xml(),
            '<div id="my_id"></div>',
        )
        self.assertEqual(
            DIV("<b>xmlscapedthis</b>").xml(),
            "<div>&lt;b&gt;xmlscapedthis&lt;/b&gt;</div>",
        )
        self.assertEqual(
            DIV(XML("<b>don'txmlscapedthis</b>")).xml(),
            "<div><b>don'txmlscapedthis</b></div>",
        )

    def test_invalid_atribute_name(self):
        i = [" ", "=", "'", '"', ">", "<", "/"]
        for x in i:
            DIV = TAG.div
            b = "_any%sthings" % x
            attr = {b: "invalid_atribute_name"}
            self.assertRaises(ValueError, DIV("any content", **attr).xml)

    def test_amend(self):
        div = DIV("hello", _class="myclass")
        div = div.amend("hello world", _id="myid")
        self.assertEqual(div.xml(), '<div class="myclass" id="myid">hello world</div>')

    def test_sanitize(self):
        permitted_tags = [
            "div",
            "td",
            "b",
            "br/",
            "strong",
            "span",
            "img/",
            "a",
        ]
        allowed_attributes = {
            "a": ["href", "title"],
            "img": ["src", "alt"],
            "blockquote": ["type"],
            "td": ["colspan"],
        }
        # test permitted
        for x in permitted_tags:
            T = TAG[x]
            s_tag = T().xml()
            if (
                x == "img/"
            ):  # alt or src attribute is required. src has to have a valid href
                s_tag = T(_alt="empty").xml()
                self.assertEqual(
                    XML(
                        s_tag,
                        sanitize=True,
                        permitted_tags=["img/"],
                        allowed_attributes={"img": ["src", "alt"]},
                    ).xml(),
                    '<img alt="empty"/>',
                )
                s_tag = T(_src="/image.png").xml()
                self.assertEqual(
                    XML(
                        s_tag,
                        sanitize=True,
                        permitted_tags=["img/"],
                        allowed_attributes={"img": ["src", "alt"]},
                    ).xml(),
                    '<img src="/image.png"/>',
                )
            elif x == "a":  # It has to have a valid href or title or not tag empty
                s_tag = T("this is a link", _href="http://web2py.com/").xml()
                self.assertEqual(
                    XML(
                        s_tag,
                        sanitize=True,
                        permitted_tags=["a"],
                        allowed_attributes={"a": ["href", "title"]},
                    ).xml(),
                    '<a href="http://web2py.com/">this is a link</a>',
                )
                s_tag = T("without href", _title="this is a link?").xml()
                self.assertEqual(
                    XML(
                        s_tag,
                        sanitize=True,
                        permitted_tags=["a"],
                        allowed_attributes={"a": ["href", "title"]},
                    ).xml(),
                    '<a title="this is a link?">without href</a>',
                )
                s_tag = T(_title="empty_tag").xml()
                self.assertEqual(
                    XML(
                        s_tag,
                        sanitize=True,
                        permitted_tags=["a"],
                        allowed_attributes={"a": ["href", "title"]},
                    ).xml(),
                    '<a title="empty_tag"></a>',
                )
            else:
                self.assertEqual(
                    XML(
                        s_tag,
                        sanitize=True,
                        permitted_tags=permitted_tags,
                        allowed_attributes=allowed_attributes,
                    ).xml(),
                    "<%s></%s>" % (x, x) if not x[-1] == "/" else "<%s>" % x,
                )

        # test tag out of list
        out_of_list = [
            "blockquote",
            "i",
            "li",
            "ol",
            "ul",
            "p",
            "cite",
            "code",
            "pre",
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
            "table",
            "tbody",
            "thead",
            "tfoot",
            "tr" "strong",
        ]
        for x in out_of_list:
            T = TAG[x]
            self.assertEqual(
                XML(
                    T().xml(),
                    sanitize=True,
                    permitted_tags=permitted_tags,
                    allowed_attributes=allowed_attributes,
                ).xml(),
                "&lt;%s&gt;&lt;/%s&gt;" % (x, x),
            )
        # test unusual tags
        for x in ["evil", "n0c1v3"]:
            T = TAG[x]
            self.assertEqual(
                XML(
                    T().xml(),
                    sanitize=True,
                    permitted_tags=permitted_tags,
                    allowed_attributes=allowed_attributes,
                ).xml(),
                "&lt;%s&gt;&lt;/%s&gt;" % (x, x),
            )
        # test allowed_attributes
        s_tag = TAG["td"]("content_td", _colspan="2", _extra_attr="invalid").xml()
        self.assertEqual(
            XML(
                s_tag,
                sanitize=True,
                permitted_tags=["td"],
                allowed_attributes={"td": ["colspan"]},
            ).xml(),
            '<td colspan="2">content_td</td>',
        )
        s_tag = TAG["a"]("link", _href="http://web2py.com/", _title="my_title").xml()
        self.assertEqual(
            XML(
                s_tag,
                sanitize=True,
                permitted_tags=["a"],
                allowed_attributes={"a": ["href", "title"]},
            ).xml(),
            '<a href="http://web2py.com/" title="my_title">link</a>',
        )
        s_tag = TAG["img/"](_alt="empty", _src="/images/logo.png").xml()
        self.assertEqual(
            XML(
                s_tag,
                sanitize=True,
                permitted_tags=["img/"],
                allowed_attributes={"img": ["src", "alt"]},
            ).xml(),
            '<img src="/images/logo.png" alt="empty"/>',
        )
        s_tag = TAG["div"]("content", _style="{backgrond-color: red;}").xml()
        self.assertEqual(
            XML(
                s_tag,
                sanitize=True,
                permitted_tags=["div"],
                allowed_attributes={"div": ["style"]},
            ).xml(),
            '<div style="{backgrond-color: red;}">content</div>',
        )
        self.assertEqual(
            XML(
                TAG["a"]("oh no!", _href="invalid_link").xml(),
                sanitize=True,
                permitted_tags=["a"],
            ).xml(),
            "oh no!",
        )
        self.assertEqual(
            XML(
                TAG["div"]("", _onclick="evil()").xml(),
                sanitize=True,
                permitted_tags=["div"],
            ).xml(),
            "<div></div>",
        )

        # valid inside invalid
        s_tag = TAG["evil"](TAG["div"]("valid"), _style="{backgrond-color: red;}").xml()
        self.assertEqual(
            XML(
                s_tag,
                sanitize=True,
                permitted_tags=["div"],
                allowed_attributes={"div": ["style"]},
            ).xml(),
            "&lt;evil&gt;<div>valid</div>&lt;/evil&gt;",
        )
        self.assertEqual(
            XML(
                TAG["a"](TAG["img/"](_src="/index.html"), _class="teste").xml(),
                sanitize=True,
                permitted_tags=["a", "img/"],
            ).xml(),
            '<img src="/index.html"/>',
        )

        # tags deleted even allowed
        self.assertEqual(
            XML(TAG["img/"]().xml(), sanitize=True, permitted_tags=["img"]).xml(), ""
        )
        self.assertEqual(
            XML(
                TAG["img/"](_src="invalid_url").xml(),
                sanitize=True,
                permitted_tags=["img"],
            ).xml(),
            "",
        )
        self.assertEqual(
            XML(
                TAG["img/"](_class="teste").xml(), sanitize=True, permitted_tags=["img"]
            ).xml(),
            "",
        )
        self.assertEqual(
            XML(
                TAG["a"](_href="invalid_link").xml(),
                sanitize=True,
                permitted_tags=["a"],
            ).xml(),
            "",
        )

    def test_find(self):
        a = DIV("A", _class="a")
        b = SPAN("B", _id="b")
        div = DIV(DIV(a), DIV(b), XML("<div class='a'>dummy</div>"))
        self.assertEqual(div.find(".a")[0], a)
        self.assertEqual(div.find("#b")[0], b)
        self.assertEqual(div.find("div.a")[0], a)
        self.assertEqual(div.find("span#b")[0], b)
        self.assertEqual(div.find("span")[0], b)
        self.assertEqual(len(div.find("div")), 4)
        self.assertEqual(len(div.find("abc")), 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
| From http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496942
| Submitter: Josh Goldfoot (other recipes)
| Last Updated: 2006/08/05
| Version: 1.0

Cross-site scripting (XSS) defense
-----------------------------------
"""

import sys
from xml.sax.saxutils import quoteattr

__all__ = ["sanitize"]

PY2 = sys.version_info[0] == 2

if PY2:
    from cgi import escape

    import urlparse
    from htmlentitydefs import entitydefs, name2codepoint
    from HTMLParser import HTMLParser
else:
    from html import escape
    from html.entities import entitydefs, name2codepoint
    from html.parser import HTMLParser
    from urllib import parse as urlparse

    basestring = str


def xmlescape(text, quote=True, colon=False):
    if not isinstance(text, basestring):
        text = str(text)
    data = escape(text, quote)
    if quote:
        data = data.replace("'", "&#x27;")
        data = data.replace('"', "&quot;")
    if colon:
        data = data.replace(":", "&#58;")
    return data


class XssCleaner(HTMLParser):
    def __init__(
        self,
        permitted_tags=[
            "a",
            "b",
            "blockquote",
            "br/",
            "i",
            "li",
            "ol",
            "ul",
            "p",
            "cite",
            "code",
            "pre",
            "img/",
        ],
        allowed_attributes={
            "a": ["href", "title"],
            "img": ["src", "alt"],
            "blockquote": ["type"],
        },
        strip_disallowed=False,
    ):

        if PY2:
            HTMLParser.__init__(self)
        else:
            HTMLParser.__init__(self, convert_charrefs=False)
        self.result = ""
        self.open_tags = []
        self.permitted_tags = [i for i in permitted_tags if i[-1] != "/"]
        self.requires_no_close = [i[:-1] for i in permitted_tags if i[-1] == "/"]
        self.permitted_tags += self.requires_no_close
        self.allowed_attributes = allowed_attributes

        # The only schemes allowed in URLs (for href and src attributes).
        # Adding "javascript" or "vbscript" to this list would not be smart.

        self.allowed_schemes = ["http", "https", "ftp", "mailto"]

        # to strip or escape disallowed tags?
        self.strip_disallowed = strip_disallowed
        # there might be data after final closing tag, that is to be ignored
        self.in_disallowed = [False]

    def handle_data(self, data):
        if data and not self.in_disallowed[-1]:
            self.result += xmlescape(data)

    def handle_charref(self, ref):
        if self.in_disallowed[-1]:
            return
        elif len(ref) < 7 and (
            ref.isdigit() or ref == "x27"
        ):  # x27 is a special case for apostrophe
            self.result += "&#%s;" % ref
        else:
            self.result += xmlescape("&#%s" % ref)

    def handle_entityref(self, ref):
        if self.in_disallowed[-1]:
            return
        elif ref in entitydefs:
            self.result += "&%s;" % ref
        else:
            self.result += xmlescape("&%s" % ref)

    def handle_comment(self, comment):
        if self.in_disallowed[-1]:
            return
        elif comment:
            self.result += xmlescape("<!--%s-->" % comment)

    def handle_starttag(self, tag, attrs):
        if tag not in self.permitted_tags:
            self.in_disallowed.append(True)
            if not self.strip_disallowed:
                self.result += xmlescape("<%s>" % tag)
        else:
            self.in_disallowed.append(False)
            bt = "<" + tag
            if tag in self.allowed_attributes:
                attrs = dict(attrs)
                self.allowed_attributes_here = [
                    x
                    for x in self.allowed_attributes[tag]
                    if x in attrs and len(attrs[x]) > 0
                ]
                for attribute in self.allowed_attributes_here:
                    if attribute in ["href", "src", "background"]:
                        if self.url_is_acceptable(attrs[attribute]):
                            bt += ' %s="%s"' % (attribute, attrs[attribute])
                    else:
                        bt += " %s=%s" % (
                            xmlescape(attribute),
                            quoteattr(attrs[attribute]),
                        )
            # deal with <a> without href and <img> without src
            if bt == "<a" or bt == "<img":
                return
            if tag in self.requires_no_close:
                bt += "/"
            bt += ">"
            self.result += bt
            if tag not in self.requires_no_close:
                self.open_tags.insert(0, tag)

    def handle_endtag(self, tag):
        bracketed = "</%s>" % tag
        self.in_disallowed and self.in_disallowed.pop()
        if tag not in self.permitted_tags:
            if not self.strip_disallowed:
                self.result += xmlescape(bracketed)
        elif tag in self.open_tags:
            self.result += bracketed
            self.open_tags.remove(tag)

    def url_is_acceptable(self, url):
        """
        Accepts relative, absolute, and mailto urls
        """

        if url.startswith("#"):
            return True
        else:
            parsed = urlparse.urlparse(url)
            return (
                (parsed[0] in self.allowed_schemes and "." in parsed[1])
                or (parsed[0] in self.allowed_schemes and "@" in parsed[2])
                or (parsed[0] == "" and parsed[2].startswith("/"))
            )

    def strip(self, rawstring, escape=True):
        """
        Returns the argument stripped of potentially harmful
        HTML or Javascript code

        @type escape: boolean
        @param escape: If True (default) it escapes the potentially harmful
          content, otherwise remove it
        """

        if not isinstance(rawstring, basestring):
            return str(rawstring)
        for tag in self.requires_no_close:
            rawstring = rawstring.replace("<%s />" % tag, "<%s/>" % tag)
        if not escape:
            self.strip_disallowed = True
        self.result = ""
        self.feed(rawstring)
        for endtag in self.open_tags:
            if endtag not in self.requires_no_close:
                self.result += "</%s>" % endtag
        return self.result

    def xtags(self):
        """
        Returns a printable string informing the user which tags are allowed
        """

        tg = ""
        for x in sorted(self.permitted_tags):
            tg += "<" + x
            if x in self.allowed_attributes:
                for y in self.allowed_attributes[x]:
                    tg += ' %s=""' % y
            tg += "> "
        return xmlescape(tg.strip())


def sanitize(
    text,
    permitted_tags=[
        "a",
        "b",
        "blockquote",
        "br/",
        "i",
        "li",
        "ol",
        "ul",
        "p",
        "cite",
        "code",
        "pre",
        "img/",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "table",
        "tbody",
        "thead",
        "tfoot",
        "tr",
        "td",
        "div",
        "strong",
        "span",
    ],
    allowed_attributes={
        "a": ["href", "title"],
        "img": ["src", "alt"],
        "blockquote": ["type"],
        "td": ["colspan"],
    },
    escape=True,
):
    if not isinstance(text, basestring):
        return str(text)
    return XssCleaner(
        permitted_tags=permitted_tags, allowed_attributes=allowed_attributes
    ).strip(text, escape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Unit tests for gluon.template
"""

import sys
import unittest

from yatl import render
from yatl.template import NOESCAPE, DummyResponse, RestrictedError


class TestTemplate(unittest.TestCase):
    def testRun(self):
        self.assertEqual(
            render(content="{{for i in range(n):}}{{=i}}{{pass}}", context=dict(n=3)),
            "012",
        )
        self.assertEqual(
            render(content="{{if n>2:}}ok{{pass}}", context=dict(n=3)), "ok"
        )
        self.assertEqual(
            render(content="{{try:}}{{n/0}}{{except:}}fail{{pass}}", context=dict(n=3)),
            "fail",
        )
        self.assertEqual(render(content='{{="<&>"}}'), "&lt;&amp;&gt;")
        self.assertEqual(render(content='"abc"'), '"abc"')
        self.assertEqual(render(content='"a\'bc"'), '"a\'bc"')
        self.assertEqual(render(content='"a"bc"'), '"a"bc"')
        self.assertEqual(render(content=r'"a\"bc"'), r'"a\"bc"')
        self.assertEqual(render(content=r'"""abc\""""'), r'"""abc\""""')

    def testEqualWrite(self):
        "test generation of response.write"
        self.assertEqual(render(content="{{=2+2}}"), "4")
        self.assertEqual(render(content='{{="abc"}}'), "abc")
        # whitespace is stripped
        self.assertEqual(render(content='{{ ="abc"}}'), "abc")
        self.assertEqual(render(content='{{ ="abc" }}'), "abc")
        self.assertEqual(render(content='{{pass\n="abc" }}'), "abc")
        # = recognized only at the beginning of a physical line
        self.assertEqual(
            render(content='{{xyz = "xyz"\n="abc"\n="def"\n=xyz }}'), "abcdefxyz"
        )
        # = in python blocks
        self.assertEqual(render(content='{{if True:\n="abc"\npass }}'), "abc")
        self.assertEqual(
            render(content='{{if True:\n="abc"\npass\n="def" }}'), "abcdef"
        )
        self.assertEqual(render(content='{{if False:\n="abc"\npass\n="def" }}'), "def")
        self.assertEqual(
            render(content='{{if True:\n="abc"\nelse:\n="def"\npass }}'), "abc"
        )
        self.assertEqual(
            render(content='{{if False:\n="abc"\nelse:\n="def"\npass }}'), "def"
        )
        # codeblock-leading = handles internal newlines, escaped or not
        self.assertEqual(render(content="{{=list((1,2,3))}}"), "[1, 2, 3]")
        self.assertEqual(render(content="{{=list((1,2,\\\n3))}}"), "[1, 2, 3]")
        self.assertEqual(render(content="{{=list((1,2,\n3))}}"), "[1, 2, 3]")
        # ...but that means no more = operators in the codeblock
        self.assertRaises(SyntaxError, render, content='{{="abc"\n="def" }}')
        # = embedded in codeblock won't handle newlines in its argument
        self.assertEqual(render(content="{{pass\n=list((1,2,\\\n3))}}"), "[1, 2, 3]")
        self.assertRaises(SyntaxError, render, content="{{pass\n=list((1,2,\n3))}}")

    def testWithDummyFileSystem(self):
        import contextlib
        from os.path import join as pjoin

        if sys.version_info[0] == 2:
            from cStringIO import StringIO
        else:
            from io import StringIO

        @contextlib.contextmanager
        def monkey_patch(module, fn_name, patch):
            try:
                unpatch = getattr(module, fn_name)
            except AttributeError:
                unpatch = None
            setattr(module, fn_name, patch)
            try:
                yield
            finally:
                if unpatch is None:
                    delattr(module, fn_name)
                else:
                    setattr(module, fn_name, unpatch)

        def dummy_open(path):
            if path == pjoin("views", "layout.html"):
                return (
                    "{{block left_sidebar}}left{{end}}"
                    "{{include}}"
                    "{{block right_sidebar}}right{{end}}"
                )
            elif path == pjoin("views", "layoutbrackets.html"):
                return (
                    "[[block left_sidebar]]left[[end]]"
                    "[[include]]"
                    "[[block right_sidebar]]right[[end]]"
                )
            elif path == pjoin("views", "default", "index.html"):
                return (
                    "{{extend 'layout.html'}}"
                    "{{block left_sidebar}}{{super}} {{end}}"
                    "to"
                    "{{block right_sidebar}} {{super}}{{end}}"
                )
            elif path == pjoin("views", "default", "indexbrackets.html"):
                return (
                    "[[extend 'layoutbrackets.html']]"
                    "[[block left_sidebar]][[super]] [[end]]"
                    "to"
                    "[[block right_sidebar]] [[super]][[end]]"
                )
            elif path == pjoin("views", "default", "missing.html"):
                return (
                    "{{extend 'wut'}}"
                    "{{block left_sidebar}}{{super}} {{end}}"
                    "to"
                    "{{block right_sidebar}} {{super}}{{end}}"
                )
            elif path == pjoin("views", "default", "noescape.html"):
                return "{{=NOESCAPE('<script></script>')}}"
            raise IOError

        self.assertEqual(
            render(
                filename=pjoin("views", "default", "index.html"),
                path="views",
                reader=dummy_open,
            ),
            "left to right",
        )
        self.assertEqual(
            render(
                filename=pjoin("views", "default", "indexbrackets.html"),
                path="views",
                delimiters="[[ ]]",
                reader=dummy_open,
            ),
            "left to right",
        )
        # self.assertRaises(
        #    RestrictedError,
        #    render,
        #    filename=pjoin('views', 'default', 'missing.html'),
        #    path='views',
        #    reader=dummy_open)
        response = DummyResponse()
        response.delimiters = ("[[", "]]")
        self.assertEqual(
            render(
                filename=pjoin("views", "default", "indexbrackets.html"),
                path="views",
                context={"response": response},
                reader=dummy_open,
            ),
            "left to right",
        )
        self.assertEqual(
            render(
                filename=pjoin("views", "default", "noescape.html"),
                context={"NOESCAPE": NOESCAPE},
                reader=dummy_open,
            ),
            "<script></script>",
        )



import copy
import copyreg as copy_reg
import marshal
import re
import sys

from . import sanitizer
from .sanitizer import xmlescape

unicodeT = unicode if sys.version_info[0] == 2 else str


def escape(value):
    if isinstance(value, str):
        return xmlescape(value)
    elif isinstance(value, bytes):
        return xmlescape(value.decode("utf8"))
    return str(value)


__all__ = [
    "A",
    "BEAUTIFY",
    "BODY",
    "CAT",
    "CODE",
    "DIV",
    "EM",
    "FORM",
    "H1",
    "H2",
    "H3",
    "H4",
    "H5",
    "H6",
    "HEAD",
    "HTML",
    "IMG",
    "INPUT",
    "LABEL",
    "LI",
    "METATAG",
    "OL",
    "OPTION",
    "P",
    "PRE",
    "SELECT",
    "SPAN",
    "STRONG",
    "TABLE",
    "TAG",
    "TAGGER",
    "THEAD",
    "TBODY",
    "TD",
    "TEXTAREA",
    "TH",
    "TT",
    "TR",
    "UL",
    "XML",
    "xmlescape",
    "escape",
    "I",
    "META",
    "LINK",
    "TITLE",
    "STYLE",
    "SCRIPT",
]

INVALID_CHARS = set(" ='\"></")


# ################################################################
# New HTML Helpers
# ################################################################


def _vk(k):
    """validate atribute name of tag
    @k: atribute name
    """
    invalid_chars = set(k) & INVALID_CHARS
    if invalid_chars:
        raise ValueError("Invalid caracters %s in attribute name" % list(invalid_chars))
    return k


class TAGGER:
    def __init__(self, name, *children, **attributes):
        self._name = name
        self.children = list(children)
        self.attributes = attributes
        for child in self.children:
            if isinstance(child, TAGGER):
                child.parent = self

    @property
    def name(self):
        return self._name

    def xml(self):
        name = self.name
        parts = []
        for key in sorted(self.attributes):
            value = self.attributes.get(key)
            if key.startswith("_") and not (value is False or value is None):
                if value is True:
                    value = _vk(key[1:])
                else:
                    value = escape(value)
                parts.append('%s="%s"' % (_vk(key[1:]), value))
        joined = " ".join(parts)
        if joined:
            joined = " " + joined
        if name.endswith("/"):
            return "<%s%s/>" % (name[0:-1], joined)
        else:
            content = "".join(
                s.xml() if is_helper(s) else escape(s) for s in self.children
            )
            return "<%s%s>%s</%s>" % (name, joined, content, name)

    def __str__(self):
        data = self.xml()
        if isinstance(data, bytes):
            data = data.decode("utf8")
        return data

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.children[key]
        else:
            return self.attributes.get(key)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.children[key] = value
        else:
            self.attributes[key] = value

    def insert(self, i, value):
        self.children.insert(i, value)

    def append(self, value):
        self.children.append(value)

    def __delitem__(self, key):
        if isinstance(key, int):
            try:
                del self.children[key]
            except IndexError:
                pass
        else:
            del self.attributes[key]

    def __len__(self):
        return len(self.children)

    def amend(self, *children, **attributes):
        new_children = children if children else copy.copy(self.children)
        new_attributes = copy.copy(self.attributes)
        new_attributes.update(**attributes)
        return TAGGER(self.name, *new_children, **new_attributes)

    regex_tag = re.compile(r"^[\w:-]+")
    regex_id = re.compile(r"#([\w-]+)")
    regex_class = re.compile(r"\.([\w-]+)")
    regex_attr = re.compile(r"\[([\w:-]+)=(.*?)\]")

    def find(self, query=None, **kargs):
        """
        Find all components that match the supplied attribute dictionary,
        or None if nothing could be found

        All components of the components are searched.

        Examples:

        >>> a = DIV(DIV(SPAN('x'),3,DIV(SPAN('y'))))
        >>> for c in a.find('span', first_only=True): c[0]='z'
        >>> print(a)
        <div><div><span>z</span>3<div><span>y</span></div></div></div>
        >>> for c in a.find('span'): c[0]='z'
        >>> print(a)
        <div><div><span>z</span>3<div><span>z</span></div></div></div>

        It also supports a syntax compatible with jQuery

        Examples:

        >>> a = DIV(SPAN(A('hello', **{'_id': '1-1', '_u:v': '$'})), P('world', _class='this is a test'))
        >>> for e in a.find('div a#1-1, p.is'): print(e)
        <a id="1-1" u:v="$">hello</a>
        <p class="this is a test">world</p>
        >>> for e in a.find('#1-1'): print(e)
        <a id="1-1" u:v="$">hello</a>
        >>> a.find('a[u:v=$]')[0].xml()
        '<a id="1-1" u:v="$">hello</a>'
        >>> a = FORM(INPUT(_type='text'),SELECT(OPTION(0)),TEXTAREA())
        >>> for c in a.find('input, select, textarea'): c['_disabled'] = True
        >>> a.xml()
        '<form><input disabled="disabled" type="text"/><select disabled="disabled"><option>0</option></select><textarea disabled="disabled"></textarea></form>'
        >>> for c in a.find('input, select, textarea'): c['_disabled'] = False
        >>> a.xml()
        '<form><input type="text"/><select><option>0</option></select><textarea></textarea></form>'

        Elements that are matched can also be replaced or removed by specifying
        a "replace" argument (note, a list of the original matching elements
        is still returned as usual).

        Examples:

        >>> a = DIV(DIV(SPAN('x', _class='abc'), DIV(SPAN('y', _class='abc'), SPAN('z', _class='abc'))))
        >>> b = a.find('span.abc', replace=P('x', _class='xyz'))
        >>> print(a)  # We should .xml() here instead of print
        <div><div><p class="xyz">x</p><div><p class="xyz">x</p><p class="xyz">x</p></div></div></div>

        "replace" can be a callable, which will be passed the original element and
        should return a new element to replace it.

        Examples:

        >>> a = DIV(DIV(SPAN('x', _class='abc'), DIV(SPAN('y', _class='abc'), SPAN('z', _class='abc'))))
        >>> b = a.find('span.abc', replace=lambda el: P(el[0], _class='xyz'))
        >>> print(a)
        <div><div><p class="xyz">x</p><div><p class="xyz">y</p><p class="xyz">z</p></div></div></div>

        If replace=None, matching elements will be removed completely.

        Examples:

        >>> a = DIV(DIV(SPAN('x', _class='abc'), DIV(SPAN('y', _class='abc'), SPAN('z', _class='abc'))))
        >>> b = a.find('span', text='y', replace=None)
        >>> print(a)
        <div><div><span class="abc">x</span><div><span class="abc"></span><span class="abc">z</span></div></div></div>

        If a "text" argument is specified, elements will be searched for text
        components that match text, and any matching text components will be
        replaced ("text" is ignored if "replace" is not also specified, use
        a "find" argument when you only need searching for textual elements).

        Like the "find" argument, "text" can be a string or a compiled regex.

        Examples:

        >>> a = DIV(DIV(SPAN('x', _class='abc'), DIV(SPAN('y', _class='abc'), SPAN('z', _class='abc'))))
        >>> b = a.find(text=re.compile('x|y|z'), replace='hello')
        >>> print(a)
        <div><div><span class="abc">hello</span><div><span class="abc">hello</span><span class="abc">hello</span></div></div></div>

        If other attributes are specified along with text, then only components
        that match the specified attributes will be searched for text.

        Examples:

        >>> a = DIV(DIV(SPAN('x', _class='abc'), DIV(SPAN('y', _class='efg'), SPAN('z', _class='abc'))))
        >>> b = a.find('span.efg', text=re.compile('x|y|z'), replace='hello')
        >>> print(a)
        <div><div><span class="abc">x</span><div><span class="efg">hello</span><span class="abc">z</span></div></div></div>
        """
        if query:
            sub = []
            if "," in query:
                # jQuery Multiple Selector ("selector1, selector2, selectorN")
                for subquery in query.split(","):
                    sub.extend(self.find(subquery.strip(), **kargs))
                return sub
            items = query.split()
            if len(items) > 1:
                # jQuery Descendant Selector ("ancestor descendant")
                subquery = " ".join(items[1:])
                for a in self.find(items[0]):
                    sub.extend(a.find(subquery, **kargs))
                return sub
            item = items[0]
            if "#" in item or "." in item or "[" in item:
                match_tag = self.regex_tag.search(item)
                match_id = self.regex_id.search(item)
                match_class = self.regex_class.search(item)
                match_attr = self.regex_attr.finditer(item)
                args = []
                if match_tag:
                    args.append(match_tag.group())
                if match_id:
                    # jQuery ID Selector ("#id")
                    kargs["_id"] = match_id.group(1)
                if match_class:
                    # jQuery Class Selector (".class")
                    kargs["_class"] = re.compile(
                        r"(?<!\w)%s(?!\w)"
                        % match_class.group(1).replace("-", r"\-").replace(":", r"\:")
                    )
                for aitem in match_attr:
                    # jQuery Attribute Equals Selector ("[name=value]")
                    kargs["_" + aitem.group(1)] = aitem.group(2)
                return self.find(*args, **kargs)
        matches = []
        if self.name is None:
            return matches
        # check if the component has an attribute with the same
        # value as provided
        tag = self.name.rstrip("/")
        # a null query found all elements
        is_matched = not query or tag == query
        for (key, value) in kargs.items():
            if key not in ("first_only", "find", "text", "replace"):
                if tag == "cat":
                    # CAT instances have not attributes
                    is_matched = False
                elif isinstance(value, (str, int)):
                    # attribute equals test (e.g. _id, _disabled)
                    if str(self[key]) != str(value):
                        is_matched = False
                elif key in self.attributes:
                    # attribute regex test (e.g. _class)
                    if not value.search(str(self[key])):
                        is_matched = False
                else:
                    is_matched = False
        if "find" in kargs:
            is_matched = False
            find = kargs["find"]
            is_regex = hasattr(find, "search")
            for c in self.children:
                if isinstance(c, str) and (
                    (is_regex and find.search(c)) or (str(find) in c)
                ):
                    is_matched = True
        # if found, return the component
        if is_matched:
            matches.append(self)

        first_only = kargs.get("first_only", False)
        replace = kargs.get("replace", False)
        text = replace is not False and kargs.get("text", False)
        is_regex = hasattr(text, "search")
        find_components = not (is_matched and first_only)

        def replace_component(i):
            if replace is None:
                del self[i]
                return i
            else:
                self[i] = replace(self[i]) if callable(replace) else replace
                return i + 1

        # loop the components
        if text or find_components:
            i = 0
            while i < len(self.children):
                c = self[i]
                j = i + 1
                if (
                    is_matched
                    and text
                    and isinstance(c, str)
                    and ((is_regex and text.search(c)) or (str(text) in c))
                ):
                    j = replace_component(i)
                elif find_components and isinstance(c, TAGGER):
                    child_matches = c.find(query, **kargs)
                    if len(child_matches):
                        if not text and replace is not False and child_matches[0] is c:
                            j = replace_component(i)
                        if first_only:
                            return child_matches
                        matches.extend(child_matches)
                i = j
        return matches


class METATAG:
    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, name):
        return lambda *children, **attributes: TAGGER(name, *children, **attributes)


class CAT(TAGGER):
    def __init__(self, *children):
        self._name = "cat"
        self.children = list(children)
        for child in self.children:
            if isinstance(child, TAGGER):
                child.parent = self

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.children[key] = value
        else:
            # attributes are set on all childrens
            for child in self.children:
                if isinstance(child, TAGGER):
                    child[key] = value

    def xml(self):
        return "".join(
            s.xml() if isinstance(s, TAGGER) else xmlescape(unicodeT(s))
            for s in self.children
        )


TAG = METATAG()
DIV = TAG.div
SPAN = TAG.span
LI = TAG.li
OL = TAG.ol
UL = TAG.ul
I = TAG.i
A = TAG.a
P = TAG.p
H1 = TAG.h1
H2 = TAG.h2
H3 = TAG.h3
H4 = TAG.h4
H5 = TAG.h5
H6 = TAG.h6
EM = TAG.em
TR = TAG.tr
TD = TAG.td
TH = TAG.th
TT = TAG.tt
PRE = TAG.pre
CODE = TAG.code
FORM = TAG.form
HEAD = TAG.head
HTML = TAG.html
BODY = TAG.body
TABLE = TAG.table
THEAD = TAG.thead
TBODY = TAG.tbody
LABEL = TAG.label
STYLE = TAG.style
STRONG = TAG.strong
SELECT = TAG.select
OPTION = TAG.option
TEXTAREA = TAG.textarea
BUTTON = TAG.button
TITLE = TAG.title
IMG = TAG["img/"]
INPUT = TAG["input/"]
META = TAG["meta/"]
LINK = TAG["link/"]
SCRIPT = lambda body, **attr: TAG.script(XML(body), **attr)


# ################################################################
# New XML Helpers
# ################################################################


class XML(TAGGER):
    """
    use it to wrap a string that contains XML/HTML so that it will not be
    escaped by the template

    Examples:

    >>> XML('<h1>Hello</h1>').xml()
    '<h1>Hello</h1>'
    """

    def __init__(
        self,
        text,
        sanitize=False,
        permitted_tags=[
            "a",
            "b",
            "blockquote",
            "br/",
            "i",
            "li",
            "ol",
            "ul",
            "p",
            "cite",
            "code",
            "pre",
            "img/",
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
            "table",
            "tr",
            "td",
            "div",
            "strong",
            "span",
        ],
        allowed_attributes={
            "a": ["href", "title", "target"],
            "img": ["src", "alt"],
            "blockquote": ["type"],
            "td": ["colspan"],
        },
    ):
        """
        Args:
            text: the XML text
            sanitize: sanitize text using the permitted tags and allowed
                attributes (default False)
            permitted_tags: list of permitted tags (default: simple list of
                tags)
            allowed_attributes: dictionary of allowed attributed (default
                for A, IMG and BlockQuote).
                The key is the tag; the value is a list of allowed attributes.
        """

        if sanitize:
            text = sanitizer.sanitize(text, permitted_tags, allowed_attributes)
        if isinstance(text, bytes):
            text = text.decode("utf8")
        self.text = text
        self._name = None

    def xml(self):
        return self.text

    def __str__(self):
        return self.text

    def __add__(self, other):
        return "%s%s" % (self, other)

    def __radd__(self, other):
        return "%s%s" % (other, self)

    def __cmp__(self, other):
        a, b = str(self), str(other)
        return (a > b) - (a < b)

    def __hash__(self):
        return hash(str(self))

    def __getitem__(self, i):
        return str(self)[i]

    def __getslice__(self, i, j):
        return str(self)[i:j]

    def __iter__(self):
        for c in str(self):
            yield c

    def __len__(self):
        return len(self.text)


def is_helper(helper):
    return hasattr(helper, "xml") and callable(helper.xml)


def XML_unpickle(data):
    return XML(marshal.loads(data))


def XML_pickle(data):
    return XML_unpickle, (marshal.dumps(str(data)),)


copy_reg.pickle(XML, XML_pickle, XML_unpickle)


# ################################################################
# BEAUTIFY everything
# ################################################################


def BEAUTIFY(obj):  # FIXME: dealing with very large objects
    if is_helper(obj):
        return obj
    elif isinstance(obj, list):
        return UL(*[LI(BEAUTIFY(item)) for item in obj])
    elif isinstance(obj, dict):
        return TABLE(
            TBODY(*[TR(TH(key), TD(BEAUTIFY(value))) for key, value in obj.items()])
        )
    elif isinstance(obj, str):
        return XML(obj)
    else:
        return repr(obj)
